<?php declare(strict_types=1);

namespace Creoline\Monitoring\Subscriber;

use Sentry\SentrySdk;
use Sentry\Tracing\SpanContext;
use Sentry\Tracing\TransactionSource;
use Shopware\Core\System\SystemConfig\SystemConfigService;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Event\RequestEvent;
use Symfony\Component\Routing\Route;
use function Sentry\continueTrace;
use function Sentry\init;

class MonitoringSubscriber implements EventSubscriberInterface
{
    /**
     * @var SystemConfigService
     */
    private $systemConfigService;

    public function __construct(SystemConfigService $systemConfigService)
    {
        $this->systemConfigService = $systemConfigService;
    }

    public static function getSubscribedEvents()
    {
        return [
            'kernel.request' => ['onKernelRequest'],
            'kernel.terminate' => ['onKernelTerminate'],
        ];
    }

    public function onKernelRequest(RequestEvent $event)
    {
        if(!$this->systemConfigService->get('CreolineMonitoring.config.active')) {
            return;
        }

        if(!$this->systemConfigService->get('CreolineMonitoring.config.dsn')) {
            return;
        }

        if($this->systemConfigService->get('CreolineMonitoring.config.traceRate')) {
            $traceRate = $this->systemConfigService->get('CreolineMonitoring.config.traceRate') / 100;
        }

        if($this->systemConfigService->get('CreolineMonitoring.config.profilingRate')) {
            $profilingRate = $this->systemConfigService->get('CreolineMonitoring.config.profilingRate') / 100;
        }

        $params = [
            'dsn' => $this->systemConfigService->get('CreolineMonitoring.config.dsn'),
            'traces_sample_rate' => $traceRate,
            'profiles_sample_rate' => $profilingRate,
        ];

        init($params);

        /** @var Request $request */
        $request = $event->getRequest();

        /** @var float $requestStartTime */
        $requestStartTime = $request->server->get('REQUEST_TIME_FLOAT', microtime(true));

        $context = continueTrace(
            $request->headers->get('sentry-trace', ''),
            $request->headers->get('baggage', '')
        );
        $context->setOp('http.server');

        $routeName = $request->attributes->get('_route');
        if(null !== $routeName && \is_string($routeName)) {
            $context->setName(sprintf('%s %s', $request->getMethod(), $routeName));
            $context->setSource(TransactionSource::route());
        } else {
            $context->setName(sprintf('%s %s%s%s', $request->getMethod(), $request->getSchemeAndHttpHost(), $request->getBaseUrl(), $request->getPathInfo()));
            $context->setSource(TransactionSource::url());
        }

        $context->setStartTimestamp($requestStartTime);
        $context->setData($this->getData($request));

        SentrySdk::getCurrentHub()->setSpan(SentrySdk::getCurrentHub()->startTransaction($context));

        $spanContext = new SpanContext();
        $spanContext->setOp('http.server');
        $spanContext->setDescription(sprintf('%s %s%s%s', $request->getMethod(), $request->getSchemeAndHttpHost(), $request->getBaseUrl(), $request->getPathInfo()));
        $spanContext->setData([
            'http.request.method' => $request->getMethod(),
            'http.url' => $request->getUri(),
            'route' => $this->getRouteName($request),
        ]);
        SentrySdk::getCurrentHub()->setSpan(SentrySdk::getCurrentHub()->getSpan()->startChild($spanContext));
    }

    public function onKernelTerminate(): void
    {
        if(!$this->systemConfigService->get('CreolineMonitoring.config.active')) {
            return;
        }

        if(!$this->systemConfigService->get('CreolineMonitoring.config.dsn')) {
            return;
        }

        $transaction = SentrySdk::getCurrentHub()->getTransaction();

        if($transaction == null) {
            return;
        }

        $transaction->finish();
    }

    protected function getRouteName(Request $request): string
    {
        $route = $request->attributes->get('_route');

        if($route instanceof Route) {
            $route = $route->getPath();
        }

        if(null === $route) {
            $route = $request->attributes->get('_controller');

            if(\is_array($route) && \is_callable($route, true)) {
                $route = sprintf('%s::%s', \is_object($route[0]) ? get_debug_type($route[0]) : $route[0], $route[1]);
            }
        }

        return \is_string($route) ? $route : '<unknown>';
    }

    private function getData(Request $request): array
    {
        $client = SentrySdk::getCurrentHub()->getClient();
        $tags = [
            'net.host.port' => (string)$request->getPort(),
            'http.request.method' => $request->getMethod(),
            'http.url' => $request->getUri(),
            'route' => $this->getRouteName($request),
        ];

        if(false !== filter_var($request->getHost(), \FILTER_VALIDATE_IP)) {
            $tags['net.host.ip'] = $request->getHost();
        } else {
            $tags['net.host.name'] = $request->getHost();
        }

        if(null !== $request->getClientIp() && null !== $client && $client->getOptions()->shouldSendDefaultPii()) {
            $tags['net.peer.ip'] = $request->getClientIp();
        }

        return $tags;
    }
}